from __future__ import absolute_import, unicode_literals, print_function

import json
import logging
import os
import os.path
import shlex

from fuzzyfinder import fuzzyfinder
from prompt_toolkit.completion import Completer, Completion

from .client import KubernetesClient
from .parser import Parser
from .utils import get_shell_option_value

logger = logging.getLogger(__name__)

# TODO: configurable
user_input_prefix_to_shell_cmd_prefix = {
    'g ': 'get ',
    'd ': 'describe ',
    'e ': 'exec -it ',
    'l ': 'logs ',
    'lt ': 'logs --tail ',
    'ld ': 'logs deploy/',
}

user_input_part_to_shell_cmd_part = {
    ' -t ': ' --tail ',
    ' -c ': ' --context ',
    ' --sort-by-age': ' --sort-by="{.metadata.creationTimestamp}"',
}

formats_to_highlight = {'yaml', 'json'}


def highlight(cmd, format):
    return cmd + '| pygmentize -l ' + format


def shell_cmd_from_user_input(user_input):
    if user_input.startswith('!'):
        return user_input[1:]

    for user_input_prefix, shell_cmd_prefix in user_input_prefix_to_shell_cmd_prefix.items():
        if user_input.startswith(user_input_prefix):
            user_input = user_input.replace(user_input_prefix, shell_cmd_prefix, 1)
            break

    for user_input_part, shell_cmd_part in user_input_part_to_shell_cmd_part.items():
        user_input = user_input.replace(user_input_part, shell_cmd_part, 1)

    output_format = get_shell_option_value(user_input, '-o', '--output')
    if output_format in formats_to_highlight:
        user_input = highlight(user_input, output_format)

    return "kubectl " + user_input


class KubectlCompleter(Completer):

    def __init__(self, suffix_to_suggestor=None):
        if suffix_to_suggestor is None:
            suffix_to_suggestor = {}
        self.end_to_suggestor = suffix_to_suggestor
        self.inline_help = True
        self.namespace = ""
        self.kube_client = KubernetesClient()

        try:
            DATA_DIR = os.path.dirname(os.path.realpath(__file__))
            DATA_PATH = os.path.join(DATA_DIR, 'data/cli.json')
            with open(DATA_PATH) as json_file:
                self.kubectl_dict = json.load(json_file)
            self.parser = Parser(DATA_PATH)
        except Exception as ex:
            logger.error("got an exception" + ex.message)

    def set_inline_help(self, val):
        self.inline_help = val

    def set_namespace(self, namespace):
        self.namespace = namespace

    def get_completions(self, document, complete_event, smart_completion=None):
        word_before_cursor = document.get_word_before_cursor(WORD=True)

        cmdline = shell_cmd_from_user_input(document.text_before_cursor.strip())
        try:
            tokens = shlex.split(cmdline)

            for suffix, suggestor in self.end_to_suggestor.items():
                if (tokens and not word_before_cursor and tokens[-1] == suffix
                        or len(tokens) > 1 and tokens[-2] == suffix):
                    for key in fuzzyfinder(word_before_cursor, suggestor()):
                        yield Completion(key, -len(word_before_cursor), display=key)
                    return

            _, _, suggestions = self.parser.parse_tokens(tokens)
            valid_keys = fuzzyfinder(word_before_cursor, suggestions.keys())
            for key in valid_keys:
                yield Completion(key, -len(word_before_cursor), display=key, display_meta=suggestions[key])
        except ValueError:
            pass



from __future__ import print_function, absolute_import

from pygments.lexer import RegexLexer
from pygments.lexer import words
from pygments.token import Literal, Text

from .completer import KubectlCompleter


class KubectlLexer(RegexLexer):
    """Provides highlighting for commands, subcommands, arguments, and options.

    """
    completer = KubectlCompleter()

    tokens = {
        'root': [
            (words(
                tuple(['kubectl', 'clear', 'exit']),
                prefix=r'\b',
                suffix=r'\b'),
             Literal.String),
            # (words(
            #     tuple(completer.all_commands),
            #     prefix=r'\b',
            #     suffix=r'\b'),
            #  Name.Class),
            # (words(
            #     tuple(completer.all_args),
            #     prefix=r'\b',
            #     suffix=r'\b'),
            #  Name.Class),
            # (words(
            #     tuple(completer.all_opts),
            #     prefix=r'',
            #     suffix=r'\b'),
            # Keyword),
            # (words(
            #     tuple(completer.global_opts),
            #     prefix=r'',
            #     suffix=r'\b'),
            # Keyword),
            # Everything else
            (r'.*\n', Text),
        ]
    }



#!/usr/bin/env python
# -*- coding: utf-8 -*-


from __future__ import print_function, absolute_import, unicode_literals

import argparse

import logging

from .utils import switch_context

logger = logging.getLogger(__name__)


def cli():
    parser = argparse.ArgumentParser()
    parser.add_argument('-c', '--context', help='The context name to use '
                                                'when starting the kube-shell.')
    args = parser.parse_args()
    if args.context:
        switch_context(args.context)

    from .app import Kubeshell
    kube_shell= Kubeshell()
    logger.info("session start")
    kube_shell.run_cli()

if __name__ == "__main__":
    cli()



from __future__ import print_function, absolute_import, unicode_literals

import logging
import os
import subprocess
import sys

import click
import yaml
from prompt_toolkit import PromptSession
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from prompt_toolkit.history import FileHistory
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.lexers import PygmentsLexer

from .client import KubernetesClient, kubeconfig_filepath
from .completer import KubectlCompleter, shell_cmd_from_user_input
from .lexer import KubectlLexer
from .style import StyleFactory
from .toolbar import Toolbar
from .utils import switch_context

logger = logging.getLogger(__name__)

inline_help = True
bindings = KeyBindings()
client = KubernetesClient()
api = client.get_core_v1_api()


class KubeConfig(object):
    clustername = user = ""
    namespace = "default"
    current_context_index = 0
    current_context_name = ""
    template_set_namespace = "kubectl config set-context {} --namespace={}"

    @staticmethod
    def parse_kubeconfig():
        if not os.path.exists(os.path.expanduser(kubeconfig_filepath)):
            return ("", "", "")

        with open(os.path.expanduser(kubeconfig_filepath), "r") as fd:
            docs = yaml.safe_load_all(fd)
            for doc in docs:
                current_context = doc.get("current-context", "")
                contexts = doc.get("contexts")
                if contexts:
                    for index, context in enumerate(contexts):
                        if context['name'] == current_context:
                            KubeConfig.current_context_index = index
                            KubeConfig.current_context_name = context['name']
                            if 'cluster' in context['context']:
                                KubeConfig.clustername = context['context']['cluster']
                            if 'namespace' in context['context']:
                                KubeConfig.namespace = context['context']['namespace']
                            if 'user' in context['context']:
                                KubeConfig.user = context['context']['user']
                            return (KubeConfig.clustername, KubeConfig.user, KubeConfig.namespace)
        return ("", "", "")

    @staticmethod
    def parse_contexts():
        if not os.path.exists(os.path.expanduser(kubeconfig_filepath)):
            return

        with open(os.path.expanduser(kubeconfig_filepath), "r") as fd:
            docs = yaml.safe_load_all(fd)
            for doc in docs:
                return doc.get("contexts")

    startup_contexts = [ctx['name'] for ctx in parse_contexts.__func__()]

    @staticmethod
    def switch_to_next_cluster():
        contexts = KubeConfig.parse_contexts()
        if contexts:
            KubeConfig.current_context_index = (KubeConfig.current_context_index + 1) % len(contexts)
            cluster_name = contexts[KubeConfig.current_context_index]['name']
            KubeConfig.switch_context(cluster_name)

    @staticmethod
    def list_namespaces():
        return client.get_resource("namespace")

    @staticmethod
    def list_namespace_names():
        return (ns[0] for ns in client.get_resource("namespace"))

    @staticmethod
    def switch_to_next_namespace(current_namespace):
        namespace_resources = KubeConfig.list_namespaces()
        namespaces = sorted(res[0] for res in namespace_resources)
        index = (namespaces.index(current_namespace) + 1) % len(namespaces)
        next_namespace = namespaces[index]
        KubeConfig.switch_namespace(next_namespace)

    @classmethod
    def switch_context(cls, ctx):
        switch_context(ctx)

    @classmethod
    def switch_namespace(cls, ns):
        kubectl_config_set_namespace = KubeConfig.template_set_namespace.format(KubeConfig.current_context_name, ns)
        cmd_process = subprocess.Popen(kubectl_config_set_namespace, shell=True, stdout=subprocess.PIPE)
        cmd_process.wait()

    # TODO: performance
    @classmethod
    def switch_context_and_namespace(cls, ctx, ns):
        kubectl_config_use_context = "kubectl config use-context " + ctx
        cmd_process = subprocess.Popen(kubectl_config_use_context, shell=True, stdout=subprocess.PIPE)
        cmd_process.wait()
        kubectl_config_set_namespace = KubeConfig.template_set_namespace.format(ctx, ns)
        cmd_process = subprocess.Popen(kubectl_config_set_namespace, shell=True, stdout=subprocess.PIPE)
        cmd_process.wait()


def make_resource_getter(resource):
    return lambda: client.get_resource(resource)


suffix_to_suggestor = {
    '-n': KubeConfig.list_namespace_names,
    '-c': lambda: KubeConfig.startup_contexts,
}

resources = [resource.name for resource in api.get_api_resources().resources]
for resource in resources:
    suffix_to_suggestor[resource] = make_resource_getter(resource)
completer = KubectlCompleter(suffix_to_suggestor)


class Kubeshell(object):
    clustername = user = ""
    namespace = "default"

    @staticmethod
    def get_message():
        prompt_style = 'class:danger' if Kubeshell.clustername == 'prod' else 'class:prompt'
        return [
            (prompt_style, '⎈  k8s:('),
            ('class:state', Kubeshell.clustername),
            (prompt_style, '/'),
            ('class:state', Kubeshell.namespace),
            (prompt_style, ') ⎈  '),
        ]

    def __init__(self, refresh_resources=True):
        shell_dir = os.path.expanduser("~/.kube/shell/")
        self.history = FileHistory(os.path.join(shell_dir, "history"))
        if not os.path.exists(shell_dir):
            os.makedirs(shell_dir)
        self.toolbar = Toolbar(KubeConfig.startup_contexts, KubeConfig.list_namespace_names,
                               self.get_cluster_name, self.get_namespace, self.get_user, self.get_inline_help)

    @bindings.add('c-x')
    def _(event):
        try:
            KubeConfig.switch_to_next_cluster()
            Kubeshell.clustername, Kubeshell.user, Kubeshell.namespace = KubeConfig.parse_kubeconfig()
        except Exception as e:
            logger.warning("failed switching clusters", exc_info=1)

    @bindings.add('c-n')
    def _(event):
        try:
            KubeConfig.switch_to_next_namespace(Kubeshell.namespace)
            Kubeshell.clustername, Kubeshell.user, Kubeshell.namespace = KubeConfig.parse_kubeconfig()
        except Exception as e:
            logger.warning("failed namespace switching", exc_info=1)

    @bindings.add('c-q')
    def _(event):
        global inline_help
        inline_help = not inline_help
        completer.set_inline_help(inline_help)

    def get_cluster_name(self):
        return Kubeshell.clustername

    def get_namespace(self):
        return Kubeshell.namespace

    def get_user(self):
        return Kubeshell.user

    def get_inline_help(self):
        return inline_help

    def update_state(self, user_input):
        ctx = client.get_context(user_input)
        ns = client.get_namespace(user_input)

        if ctx and ctx != Kubeshell.clustername and ns and ns != Kubeshell.namespace:
            KubeConfig.switch_context_and_namespace(ctx, ns)
            return
        if ctx and ctx != Kubeshell.clustername:
            KubeConfig.switch_context(ctx)
            return
        if ns and ns != Kubeshell.namespace:
            KubeConfig.switch_namespace(ns)

    def run_cli(self):

        logger.info("running kube-shell event loop")
        if not os.path.exists(os.path.expanduser(kubeconfig_filepath)):
            click.secho('Kube-shell uses {0} for server side completion. Could not find {0}. '
                        'Server side completion functionality may not work.'.format(kubeconfig_filepath),
                        fg='red', blink=True, bold=True)

        session = PromptSession(history=self.history)

        while True:
            global inline_help
            try:
                Kubeshell.clustername, Kubeshell.user, Kubeshell.namespace = KubeConfig.parse_kubeconfig()
            except:
                logger.error("unable to parse {} %s".format(kubeconfig_filepath), exc_info=1)
            completer.set_namespace(self.namespace)

            try:
                user_input = session.prompt(
                    self.get_message,
                    auto_suggest=AutoSuggestFromHistory(),
                    style=StyleFactory("vim").style,
                    lexer=PygmentsLexer(KubectlLexer),
                    enable_history_search=False,
                    bottom_toolbar=self.toolbar.handler,
                    key_bindings=bindings,
                    completer=completer,
                    complete_in_thread=True
                )
            except (EOFError, KeyboardInterrupt):
                sys.exit()

            if user_input == "clear":
                click.clear()
            elif user_input == "exit":
                sys.exit()

            user_input = shell_cmd_from_user_input(user_input)

            if user_input:
                p = subprocess.Popen(user_input, shell=True)
                try:
                    _, stderr_data = p.communicate()
                    if not stderr_data:
                        self.update_state(user_input)
                except KeyboardInterrupt:
                    print()



import shlex
import subprocess


def get_shell_option_value(cmd, *options):
    found = False
    try:
        tokens = shlex.split(cmd)
    except ValueError:
        return
    for token in tokens:
        if found:
            return token
        if token in '<|>':
            return
        if token in options:
            found = True


def switch_context(ctx):
    kubectl_config_use_context = "kubectl config use-context " + ctx
    cmd_process = subprocess.Popen(kubectl_config_use_context, shell=True, stdout=subprocess.PIPE)
    cmd_process.wait()



from __future__ import absolute_import, unicode_literals, print_function

import logging
import os

import urllib3
from kubernetes import client, config
from kubernetes.client.api_client import ApiException
from urllib3.exceptions import NewConnectionError, ConnectTimeoutError, MaxRetryError

from .utils import get_shell_option_value

# disable warnings on stdout/stderr from urllib3 connection errors
ulogger = logging.getLogger("urllib3")
ulogger.setLevel("ERROR")
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
kubeconfig_filepath = os.getenv("KUBECONFIG") or "~/.kube/config"


class KubernetesClient(object):

    def __init__(self):
        self.logger = logging.getLogger(__name__)
        try:
            config_file = os.path.expanduser(kubeconfig_filepath)
            config.load_kube_config(config_file=config_file)
        except:
            self.logger.warning("unable to load kube-config")

        self.v1 = client.CoreV1Api()
        self.v1Beta1 = client.AppsV1beta1Api()
        self.extensionsV1Beta1 = client.ExtensionsV1beta1Api()
        self.autoscalingV1Api = client.AutoscalingV1Api()
        self.rbacApi = client.RbacAuthorizationV1beta1Api()
        self.batchV1Api = client.BatchV1Api()
        self.batchV2Api = client.BatchV2alpha1Api()

    def get_resource(self, resource, namespace="all"):
        ret, resources = None, list()
        try:
            ret, namespaced_resource = self._call_api_client(resource)
        except ApiException as ae:
            self.logger.warning("resource autocomplete disabled, encountered "
                                "ApiException", exc_info=1)
        except (NewConnectionError, MaxRetryError, ConnectTimeoutError):
            self.logger.warning("unable to connect to k8 cluster", exc_info=1)
        if ret:
            for i in ret.items:
                if namespace == "all" or not namespaced_resource:
                    resources.append((i.metadata.name, i.metadata.namespace))
                elif namespace == i.metadata.namespace:
                    resources.append((i.metadata.name, i.metadata.namespace))
        return resources

    def _call_api_client(self, resource):
        namespaced_resource = True

        if resource == "pod":
            ret = self.v1.list_pod_for_all_namespaces(watch=False)
        elif resource == "service":
            ret = self.v1.list_service_for_all_namespaces(watch=False)
        elif resource == "deployment":
            ret = self.v1Beta1.list_deployment_for_all_namespaces(watch=False)
        elif resource == "statefulset":
            ret = self.v1Beta1.list_stateful_set_for_all_namespaces(watch=False)
        elif resource == "node":
            namespaced_resource = False
            ret = self.v1.list_node(watch=False)
        elif resource == "namespace":
            namespaced_resource = False
            ret = self.v1.list_namespace(watch=False)
        elif resource == "daemonset":
            ret = self.extensionsV1Beta1.list_daemon_set_for_all_namespaces(watch=False)
        elif resource == "networkpolicy":
            ret = self.extensionsV1Beta1.list_network_policy_for_all_namespaces(watch=False)
        elif resource == "thirdpartyresource":
            namespaced_resource = False
            ret = self.extensionsV1Beta1.list_third_party_resource(watch=False)
        elif resource == "replicationcontroller":
            ret = self.v1.list_replication_controller_for_all_namespaces(watch=False)
        elif resource == "replicaset":
            ret = self.extensionsV1Beta1.list_replica_set_for_all_namespaces(watch=False)
        elif resource == "ingress":
            ret = self.extensionsV1Beta1.list_ingress_for_all_namespaces(watch=False)
        elif resource == "endpoints":
            ret = self.v1.list_endpoints_for_all_namespaces(watch=False)
        elif resource == "configmap":
            ret = self.v1.list_config_map_for_all_namespaces(watch=False)
        elif resource == "event":
            ret = self.v1.list_event_for_all_namespaces(watch=False)
        elif resource == "limitrange":
            ret = self.v1.list_limit_range_for_all_namespaces(watch=False)
        elif resource == "configmap":
            ret = self.v1.list_config_map_for_all_namespaces(watch=False)
        elif resource == "persistentvolume":
            namespaced_resource = False
            ret = self.v1.list_persistent_volume(watch=False)
        elif resource == "secret":
            ret = self.v1.list_secret_for_all_namespaces(watch=False)
        elif resource == "resourcequota":
            ret = self.v1.list_resource_quota_for_all_namespaces(watch=False)
        elif resource == "componentstatus":
            namespaced_resource = False
            ret = self.v1.list_component_status(watch=False)
        elif resource == "podtemplate":
            ret = self.v1.list_pod_template_for_all_namespaces(watch=False)
        elif resource == "serviceaccount":
            ret = self.v1.list_service_account_for_all_namespaces(watch=False)
        elif resource == "horizontalpodautoscaler":
            ret = self.autoscalingV1Api.list_horizontal_pod_autoscaler_for_all_namespaces(watch=False)
        elif resource == "clusterrole":
            namespaced_resource = False
            ret = self.rbacApi.list_cluster_role(watch=False)
        elif resource == "clusterrolebinding":
            namespaced_resource = False
            ret = self.rbacApi.list_cluster_role_binding(watch=False)
        elif resource == "job":
            ret = self.batchV1Api.list_job_for_all_namespaces(watch=False)
        elif resource == "cronjob":
            ret = self.batchV2Api.list_cron_job_for_all_namespaces(watch=False)
        elif resource == "scheduledjob":
            ret = self.batchV2Api.list_scheduled_job_for_all_namespaces(watch=False)
        else:
            return None, namespaced_resource
        return ret, namespaced_resource

    def get_core_v1_api(self):
        return self.v1

    @staticmethod
    def get_context(cmd):
        return get_shell_option_value(cmd, '-c', '--context')

    @staticmethod
    def get_namespace(cmd):
        return get_shell_option_value(cmd, '-n', '--namespace')



from kubeshell import __version__
from setuptools import setup, find_packages

import sys

version = sys.version_info
error_msg = "kube-shell needs Python>=2.7.10. Found %s" % sys.version

if version.major == 2:
    if version.minor < 7:
        sys.exit(error_msg)
    else:
        if version.micro < 10:
            sys.exit(error_msg)


requires = [
    'prompt-toolkit==2.0.2',
    'Pygments>=2.1.3,<3.0.0',
    'fuzzyfinder>=1.0.0',
    'click>=4.0,<7.0',
    'kubernetes>=0.10.0,<3.0.0',
]

setup(
    name='kbsh',
    version=__version__,
    description='An integrated shell for working effectively with multiple Kubernetes clusters',
    long_description=open('README.rst').read(),
    author='Lei Zhang',
    url='https://github.com/leizhag/kbsh',
    packages=find_packages(),
    package_data={'kubeshell': ['data/cli.json']},
    zip_safe=False,
    install_requires=requires,
    entry_points={
        'console_scripts': [
            'kbsh = kubeshell.main:cli',
        ]
    },
    license="Apache License 2.0",
    keywords=('kubernetes', 'autocomplete', 'shell',),
    classifiers=(
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Natural Language :: English',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ),
)



from __future__ import absolute_import, unicode_literals, print_function

import json
import logging

logger = logging.getLogger(__name__)

from .client import KubernetesClient


class Option(object):
    """ Option represents an optional local flag in kubectl """

    def __init__(self, name, helptext):
        self.name = name
        self.helptext = helptext


class CommandTree(object):
    """ CommandTree represents the tree node of a kubectl command line """

    def __init__(self, node=None, helptext=None, children=None, localFlags=None):
        self.node = node
        self.help = helptext
        self.children = children if children else list()
        self.localFlags = localFlags if localFlags else list()

    def __str__(self):
        return "Node: %s, Help: %s\n Flags: %s\n Children: %s" % (self.node, self.help, self.localFlags, self.children)


class Parser(object):
    """ Parser builds and walks the syntax tree of kubectl """

    def __init__(self, apiFile):
        self.json_api = apiFile
        self.schema = dict()
        self.globalFlags = list()
        with open(self.json_api) as api:
            self.schema = json.load(api)
        self.ast = CommandTree("kubectl")
        self.ast = self.build(self.ast, self.schema.get("kubectl"))
        self.kube_client = KubernetesClient()

    def build(self, root, schema):
        """ Build the syntax tree for kubectl command line """
        if schema.get("subcommands") and schema["subcommands"]:
            for subcmd, childSchema in schema["subcommands"].items():
                child = CommandTree(node=subcmd)
                child = self.build(child, childSchema)
                root.children.append(child)
        # {args: {}, options: {}, help: ""}
        root.help = schema.get("help")
        for name, desc in schema.get("options").items():
            if root.node == "kubectl":  # register global flags
                self.globalFlags.append(Option(name, desc["help"]))
            root.localFlags.append(Option(name, desc["help"]))
        for arg in schema.get("args"):
            node = CommandTree(node=arg)
            root.children.append(node)
        return root

    def print_tree(self, root, indent=0):
        indentter = '{:>{width}}'.format(root.node, width=indent)
        print(indentter)
        for child in root.children:
            self.print_tree(root=child, indent=indent + 2)

    def parse_tokens(self, tokens):
        """ Parse a sequence of tokens

        returns tuple of (parsed tokens, suggestions)
        """
        if len(tokens) == 1:
            return list(), tokens, {"kubectl": self.ast.help}
        else:
            tokens.reverse()
        parsed, unparsed, suggestions = self.treewalk(self.ast, parsed=list(), unparsed=tokens)
        if not suggestions and unparsed:
            # TODO: @vogxn: This is hack until we include expected value types for each option and argument.
            # Whenver we recieve no suggestions but are left with unparsed tokens we pop the value and walk the
            # tree again without values
            logger.debug("unparsed tokens remain, possible value encountered")
            unparsed.pop()
            parsed.reverse()
            unparsed.extend(parsed)
            logger.debug("resuming treewalk with tokens: %s", unparsed)
            return self.treewalk(self.ast, parsed=list(), unparsed=unparsed)
        else:
            return parsed, unparsed, suggestions

    def treewalk(self, root, parsed, unparsed):
        """ Recursively walks the syntax tree at root and returns
        the items parsed, unparsed and possible suggestions """
        suggestions = dict()
        if not unparsed:
            logger.debug("no tokens left unparsed. returning %s, %s", parsed, suggestions)
            return parsed, unparsed, suggestions

        token = unparsed.pop().strip()
        logger.debug("begin parsing at %s w/ tokens: %s", root.node, unparsed)
        if root.node == token:
            logger.debug("root node: %s matches next token:%s", root.node, token)
            parsed.append(token)
            if self.peekForOption(unparsed):  # check for localFlags and globalFlags
                logger.debug("option(s) upcoming %s", unparsed)
                parsed_opts, unparsed, suggestions = self.evalOptions(root, list(), unparsed[:])
                if parsed_opts:
                    logger.debug("parsed option(s): %s", parsed_opts)
                    parsed.extend(parsed_opts)
            if unparsed and not self.peekForOption(unparsed):  # unparsed bits without options
                logger.debug("begin subtree %s parsing", root.node)
                for child in root.children:
                    parsed_subtree, unparsed, suggestions = self.treewalk(child, list(), unparsed[:])
                    if parsed_subtree:  # subtree returned further parsed tokens
                        parsed.extend(parsed_subtree)
                        logger.debug("subtree at: %s has matches. %s, %s", child.node, parsed, unparsed)
                        break
                else:
                    # no matches found in command tree
                    # return children of root as suggestions
                    logger.debug("no matches in subtree: %s. returning children as suggestions", root.node)
                    for child in root.children:
                        suggestions[child.node] = child.help
        else:
            logger.debug("no token or option match")
            unparsed.append(token)
        return parsed, unparsed, suggestions

    def peekForOption(self, unparsed):
        """ Peek to find out if next token is an option """
        if unparsed and unparsed[-1].startswith("--"):
            return True
        return False

    def evalOptions(self, root, parsed, unparsed):
        """ Evaluate only the options and return flags as suggestions """
        logger.debug("parsing options at tree: %s with p:%s, u:%s", root.node, parsed, unparsed)
        suggestions = dict()
        token = unparsed.pop().strip()

        parts = token.partition('=')
        if parts[-1] != '':  # parsing for --option=value type input
            token = parts[0]

        allFlags = root.localFlags + self.globalFlags
        for flag in allFlags:
            if flag.name == token:
                logger.debug("matched token: %s with flag: %s", token, flag.name)
                parsed.append(token)
                if self.peekForOption(unparsed):  # recursively look for further options
                    parsed, unparsed, suggestions = self.evalOptions(root, parsed, unparsed[:])
                # elif token == "--namespace":
                #     namespaces = [('default', None), ('minikube', None), ('gitlab', None)]  # self.kube_client.get_resource("namespace")
                #     suggestions = dict(namespaces)
                break
        else:
            logger.debug("no flags match, returning allFlags suggestions")
            for flag in allFlags:
                suggestions[flag.name] = flag.helptext

        if suggestions:  # incomplete parse, replace token
            logger.debug("incomplete option: %s provided. returning suggestions", token)
            unparsed.append(token)
        return parsed, unparsed, suggestions


if __name__ == '__main__':
    parser = Parser('/Users/tsp/workspace/py/kube-shell/kubeshell/data/cli.json')
    p, _, s = parser.treewalk(parser.ast, parsed=list(),
                              unparsed=['--', '--tcp 900:8080', 'nodeport', 'service', 'create', 'kubectl'])
    print(p, s)



from __future__ import print_function, absolute_import, unicode_literals
from pygments.token import Token
from pygments.token import Keyword, Name, Operator, Generic, Literal, Text

class Toolbar(object):
    """Show information about the aws-shell in a tool bar.

    :type handler: callable
    :param handler: Wraps the callable `get_toolbar_items`.

    """

    def __init__(self, contexts, list_namespaces, get_cluster_name, get_namespace, get_user, get_inline_help):
        self.handler = self._create_toolbar_handler(contexts, list_namespaces, get_cluster_name,
                                                    get_namespace, get_user, get_inline_help)

    def _create_toolbar_handler(self, contexts, list_namespaces, get_cluster_name,
                                get_namespace, get_user, get_inline_help):
        def get_toolbar_items():
            if get_inline_help():
                help_token = 'class:bottom-toolbar.on'
                help = "ON"
            else:
                help_token = 'class:bottom-toolbar.off'
                help = "OFF"

            return [
                ('class:bottom-toolbar', 'Contexts: ' + ' '.join(contexts) + '\n'),
                ('class:bottom-toolbar', 'Namespaces: ' + ' '.join(list_namespaces()) + '\n'),
                ('class:keyword',        '[C-X] switch-context  '),
                ('class:keyword',        '[C-N] swith-namespace  '),
                ('class:keyword',        '[C-H] in-line help: '),
                (help_token,             '{0}  '.format(help)),
                ('class:keyword',        '[C-C] exit')
            ]

        return get_toolbar_items
